#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "get_date.h"
#include "ptfps.h"
#include "ptfps_linux.h"
#include "debug.h"

#define PSCMD	"ps --cols 4095 -A -o ruid -o uid -o uid -o rgid -o gid -o gid -o ruser -o user -o user -o rgroup -o group -o group -o pid -o ppid -o stime -o time -o pcpu -o pmem -o vsz -o rss -o flags -o pri -o ni -o s -o wchan -o tty -o comm -o args"

#define MAX_LINE_LEN	4096
#define MAX_TIME_LEN	16
#define MIN(x, y)       ((x) < (y) ? (x) : (y))
#define WHITESPACE		" \t\n\v\f\r"

static time_t hr_min_sec_to_time_t(char *hms)
{
    int h, m, s;

    sscanf(hms, "%d:%d:%d", &h, &m, &s);

    return (time_t)(s + (m * 60) + (h * 3600));
}

process_t *get_next_process()
{
    static FILE *f = NULL;
    static char line[MAX_LINE_LEN];
    static char stime[MAX_TIME_LEN], cputime[MAX_TIME_LEN];
    static process_t pr;
    int offset, len;
    time_t t;
    char state[2];

    if (f == NULL){
        f = popen(PSCMD, "r");
        if (f == NULL){
            pt_debug("Unable to open pipe to ps\n");
            return NULL;
        }

        if (fgets(line, MAX_LINE_LEN, f) == NULL){
            pt_debug("Unable to get header from ps\n");
            pclose(f);
            return NULL;
        }

        pt_debug("line = '%s'\n", line);
    }

    if (fgets(line, MAX_LINE_LEN, f) == NULL){
        pt_debug("line = '%s'\n", line);
        pt_debug("Got last line\n");
        pclose(f);
        return NULL;
    }

    sscanf(line, "%d %d %d"  /* uid's */
            "%d %d %d"  /* gid's */
            "%16s %16s %16s" /* usernames */
            "%16s %16s %16s" /* groupnames */
            "%d %d" /* pid's */
            "%16s %16s" /* time's */
            "%lf %lf" /* cpu and memory */
            "%lu %lu %lo" /* vsz rss flags */
            "%d %d" /* pri nice */
            "%1s" /* state */
            "%10s" /* wchan */
            "%32s" /* tty */
            "%16s" /* cmd */
            "%n" /* args */,
            &pr.ruid, &pr.euid, &pr.suid,
            &pr.rgid, &pr.egid, &pr.sgid,
            pr.ruser, pr.euser, pr.suser,
            pr.rgroup, pr.egroup, pr.sgroup,
            &pr.pid, &pr.ppid,
            stime, cputime,
            &pr.pcpu, &pr.pmem,
            &pr.vsz, &pr.rss, &pr.flags,
            &pr.pri, &pr.nice,
            state,
            pr.wchan,
            pr.tty,
            pr.cmd,
            &offset);

            pr.state = state[0];
            /* Strip the whitespace from the beginning of the command */
            offset += strspn(line+offset, WHITESPACE);
            /* Use strncat to ensure null-termination */
            pr.args[0] = '\0';
            strncat(pr.args, line+offset, 80);
            /* Trim the newline */
            len = strlen(pr.args)-1;
            if (pr.args[len] == '\n')
                pr.args[len] = '\0';

            time(&t);
            pr.stime = get_date(stime, &t);
            pr.cputime = hr_min_sec_to_time_t(cputime);

            return &pr;
}

/* UID's */
uid_t get_ruid(process_t *p)
{
    return p->ruid;
}
uid_t get_euid(process_t *p)
{
    return p->euid;
}
uid_t get_suid(process_t *p)
{
    return p->suid;
}

/* GID's */
gid_t get_rgid(process_t *p)
{
    return p->rgid;
}

gid_t get_egid(process_t *p)
{
    return p->egid;
}

gid_t get_sgid(process_t *p)
{
    return p->sgid;
}

/* Username's */
char *get_rusername(process_t *p)
{
    return p->ruser;
}

char *get_eusername(process_t *p)
{
    return p->euser;
}

char *get_susername(process_t *p)
{
    return p->suser;
}

/* Group names */
char *get_rgroupname(process_t *p)
{
    return p->rgroup;
}

char *get_egroupname(process_t *p)
{
    return p->egroup;
}

char *get_sgroupname(process_t *p)
{
    return p->sgroup;
}

/* Etc */
char *get_name(process_t *p)
{
    return p->args;
}

char *get_cmd(process_t *p)
{
    return p->cmd;
}

status_t get_state(process_t *p)
{
    switch (p->state) {
        case 'R':
            return ST_RUNNABLE;
        case 'D':
        case 'S':
            return ST_SLEEPING;
        case 'T':
            return ST_STOPPED;
        case 'Z':
            return ST_ZOMBIE;
        default:
            return ST_UNKNOWN;
    }
}

char *get_wchan(process_t *p)
{
    return p->wchan;
}

char *get_tty(process_t *p)
{
    return p->tty;
}

pid_t get_pid(process_t *p)
{
    return p->pid;
}

pid_t get_ppid(process_t *p)
{
    return p->ppid;
}

time_t get_stime(process_t *p)
{
    return p->stime;
}

unsigned long get_time(process_t *p)
{
    return p->cputime;
}

double get_pcpu(process_t *p)
{
    return p->pcpu;
}

double get_pmem(process_t *p)
{
    return p->pmem;
}

unsigned long get_rss(process_t *p)
{
    return p->rss;
}

unsigned long get_vsz(process_t *p)
{
    return p->vsz;
}

unsigned long get_flags(process_t *p)
{
    return p->flags;
}

int get_pri(process_t *p)
{
    return p->pri;
}

int get_nice(process_t *p)
{
    return p->nice;
}
